from bpm_ai_core.llm.common.llm import LLM
from bpm_ai_core.ocr.ocr import OCR
from bpm_ai_core.prompt.prompt import Prompt
from bpm_ai_core.speech_recognition.asr import ASRModel
from bpm_ai_core.tracing.decorators import trace
from bpm_ai_core.translation.nmt import NMTModel

from bpm_ai.common.errors import MissingParameterError, LanguageNotFoundError
from bpm_ai.common.multimodal import transcribe_audio, prepare_images_for_llm_prompt, ocr_documents, prepare_text_blobs, \
    assert_all_files_processed, replace_text_blobs
from bpm_ai.translate.util import get_translation_output_schema, get_lang_code


@trace("bpm-ai-translate", ["llm"])
async def translate_llm(
        llm: LLM,
        input_data: dict[str, str | dict | None],
        target_language: str,
        ocr: OCR | None = None,
        asr: ASRModel | None = None
) -> dict:
    input_items = {k: v for k, v in input_data.items() if v is not None}
    if not input_items:
        return input_data

    if not target_language or target_language.isspace():
        raise MissingParameterError("target language is required")

    if not ocr and llm.supports_images():
        input_items = prepare_images_for_llm_prompt(input_items)
    else:
        input_items = await ocr_documents(input_items, ocr)
    input_items = await transcribe_audio(input_items, asr)
    input_items = prepare_text_blobs(input_items)
    assert_all_files_processed(input_items)

    prompt = Prompt.from_file(
        "translate",
        input=input_items,
        lang=target_language
    )

    compose_schema = {
        "name": "store_translation",
        "description": f"Stores the finished translation into {target_language}.",
        "type": "object",
        "properties": get_translation_output_schema(input_items, target_language)
    }

    message = await llm.generate_message(prompt, output_schema=compose_schema)

    return {k: message.content.get(k, None) for k in input_data.keys()}


@trace("bpm-ai-translate", ["nmt"])
async def translate_nmt(
        nmt: NMTModel,
        input_data: dict[str, str | dict | None],
        target_language: str,
        ocr: OCR | None = None,
        asr: ASRModel | None = None
) -> dict:
    input_items = {k: v for k, v in input_data.items() if v is not None}
    if not input_items:
        return input_data

    if not target_language or target_language.isspace():
        raise MissingParameterError("target language is required")

    input_items = await ocr_documents(input_items, ocr)
    input_items = await transcribe_audio(input_items, asr)
    input_items = await replace_text_blobs(input_items)
    assert_all_files_processed(input_items)

    try:
        target_language_code = get_lang_code(target_language)
    except LookupError:
        raise LanguageNotFoundError(f"Could not identify target language '{target_language}'.")

    texts_to_translate = list(input_items.values())
    texts_translated = await nmt.translate(texts_to_translate, target_language_code)
    input_items_translated = {k: texts_translated[i] for i, k in enumerate(input_items.keys())}

    return {k: input_items_translated.get(k, None) for k in input_data.keys()}



from enum import Enum


def get_translation_output_schema(input_data: dict, target_language: str):
    return {k: f'{k} translated into {target_language}' for k in input_data.keys()}


class Language(Enum):
    ENGLISH = "en"
    FRENCH = "fr"
    GERMAN = "de"
    SPANISH = "es"
    ITALIAN = "it"
    PORTUGUESE = "pt"
    DUTCH = "nl"
    DANISH = "da"
    SWEDISH = "sv"
    NYNORSK = "no"
    NORWEGIAN = "no"
    BOKMAL = "no"
    FINNISH = "fi"
    POLISH = "pl"
    UKRAINIAN = "uk"


language_dict = {lang.lower(): Language[lang].value for lang in Language.__members__.keys()}


def get_lang_code(language: str) -> str:
    return language_dict[language.strip().lower()]


import json
import re
from typing import TypedDict, Callable

from bpm_ai_core.llm.common.llm import LLM
from bpm_ai_core.llm.common.message import UserMessage
from bpm_ai_core.ocr.ocr import OCR
from bpm_ai_core.prompt.prompt import Prompt
from bpm_ai_core.speech_recognition.asr import ASRModel
from bpm_ai_core.tracing.decorators import trace

from bpm_ai.common.errors import MissingParameterError
from bpm_ai.common.multimodal import transcribe_audio, prepare_images_for_llm_prompt, ocr_documents, prepare_text_blobs, \
    assert_all_files_processed
from bpm_ai.compose.util import remove_stop_words, type_to_prompt_type_str, desc_to_var_name

TEMPLATE_VAR_PATTERN = r'\{\s*([^{}\s]+(?:\s*[^{}\s]+)*)\s*\}'


class TextProperties(TypedDict):
    style: str
    type: str
    tone: str
    length: str
    language: str
    temperature: str


@trace("bpm-ai-compose", ["llm"])
async def compose_llm(
    llm: LLM,
    input_data: dict[str, str | dict],
    template: str,
    properties: TextProperties,
    ocr: OCR | None = None,
    asr: ASRModel | None = None
) -> dict:
    if template is None:
        raise MissingParameterError("template is required")

    def format_vars(template: str, f: Callable[[str], str]):
        return re.sub(TEMPLATE_VAR_PATTERN, lambda m: f(m.group(1)), template)

    if not ocr and llm.supports_images():
        input_data = prepare_images_for_llm_prompt(input_data)
    else:
        input_data = await ocr_documents(input_data, ocr)
    input_data = await transcribe_audio(input_data, asr)
    input_data = prepare_text_blobs(input_data)
    assert_all_files_processed(input_data)

    # all variables found in the template
    template_vars = re.findall(TEMPLATE_VAR_PATTERN, template)
    # map of template variables that are not already in the input and need to be generated
    template_vars_to_generate_dict = {desc_to_var_name(desc): desc for desc in template_vars if
                                      desc not in input_data.keys()}
    # input variables that are not present in the template
    non_template_input_var_dict = {k: v for k, v in input_data.items() if k not in template_vars}

    if len(template_vars_to_generate_dict) > 0:
        prompt = Prompt.from_file(
            "compose",
            context=non_template_input_var_dict,
            # remove template braces from variables that are already present in the input to not confuse the model what to generate
            # we do not resolve these variables here to avoid sending that data to the API
            template=format_vars(template, lambda v: v if v in input_data.keys() else '{' + v + '}'),
            type=type_to_prompt_type_str(properties.get("type", "letter")),
            style=properties.get("style", "formal"),
            tone=properties.get("tone", "friendly"),
            length=properties.get("length", "adequate"),
            lang=properties.get("language", "English")
        )

        compose_schema = {
            "name": "store_text",
            "description": "Stores composed text parts for template variables.",
            "type": "object",
            "properties": template_vars_to_generate_dict
        }

        message = await llm.generate_message(prompt, output_schema=compose_schema)

        generated_vars = message.content or {}
    else:
        generated_vars = {}

    input_vars = {desc_to_var_name(k): v for k, v in input_data.items()}
    all_vars = generated_vars | input_vars

    # resolve all template variables using either input or generated values
    result = format_vars(template, lambda v: all_vars[desc_to_var_name(v)])

    return {"text": result}







def get_decision_output_schema(
    output_type: str,
    possible_values: list | None = None,
    multiple_decision_values: bool = False
):
    return {
        "0_reasoning": "Concise description of the reasoning behind the decision",
        "1_decision": {
            "description": f"The final decision value{'s' if multiple_decision_values else ''}, may be null if no decision was possible",
            **({"type": output_type, **({"enum": possible_values} if possible_values is not None else {})}
               if not multiple_decision_values else _get_multi_value_schema(output_type, possible_values))
        }
    }


def get_cot_decision_output_schema(
    output_type: str,
    possible_values: list | None = None,
    multiple_decision_values: bool = False
):
    return {
        "0_reasoning": {
            "type": "object",
            "properties": {
                "00_relevantFacts": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "description": "A discrete fact"
                    }
                },
                "01_deducedInformation": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "description": "Additional information that can be deduced from the relevantFacts"
                    }
                },
                "02_reasoningSteps": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "description": "A discrete reasoning step. Do not perform multiple steps in one. Be very fine-grained and use discrete steps/items."
                    }
                },
                "03_finalReasoning": {
                    "type": "string",
                    "description": "Concise description of the final reasoning behind the decision"
                }
            }
        },
        "1_decision": {
            "description": f"The final decision value{'s' if multiple_decision_values else ''}, may be null if no decision was possible",
            **({"type": output_type, **({"enum": possible_values} if possible_values is not None else {})}
               if not multiple_decision_values else _get_multi_value_schema(output_type, possible_values))
        }
    }


def _get_multi_value_schema(output_type, possible_values):
    multi_value_schema = {
        "type": "array",
        "items": {
            "type": output_type,
            "description": "A single decision value",
            **({"enum": possible_values} if possible_values is not None else {})
        }
    }
    return multi_value_schema


def remove_order_prefix_from_keys(data):
    if isinstance(data, dict):
        return {key.lstrip('0123456789_'): remove_order_prefix_from_keys(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [remove_order_prefix_from_keys(item) for item in data]
    else:
        return data



from bpm_ai_core.llm.common.llm import LLM
from bpm_ai_core.ocr.ocr import OCR
from bpm_ai_core.prompt.prompt import Prompt
from bpm_ai_core.speech_recognition.asr import ASRModel
from bpm_ai_core.tracing.decorators import trace

from bpm_ai.common.errors import MissingParameterError
from bpm_ai.common.multimodal import transcribe_audio, prepare_images_for_llm_prompt, ocr_documents, prepare_text_blobs, \
    assert_all_files_processed


@trace("bpm-ai-generic", ["llm"])
async def generic_llm(
    llm: LLM,
    input_data: dict[str, str | dict],
    instructions: str,
    output_schema: dict[str, str | dict],
    ocr: OCR | None = None,
    asr: ASRModel | None = None
) -> dict:
    if not instructions or instructions.isspace():
        raise MissingParameterError("instructions are required")
    if not output_schema:
        raise MissingParameterError("output schema is required")

    if not ocr and llm.supports_images():
        input_data = prepare_images_for_llm_prompt(input_data)
    else:
        input_data = await ocr_documents(input_data, ocr)
    input_data = await transcribe_audio(input_data, asr)
    input_data = prepare_text_blobs(input_data)
    assert_all_files_processed(input_data)

    prompt = Prompt.from_file(
        "generic",
        context=input_data,
        task=instructions,
    )

    generic_schema = {
        "name": "store_task_result",
        "description": "Stores the result of the task.",
        "type": "object",
        "properties": output_schema
    }

    message = await llm.generate_message(prompt, output_schema=generic_schema)

    return message.content or {}



